TyNum = int | float
TnInt = None | int
TnNum = None | TyNum


class Num:
    """ Manage Number Class
    """
    @staticmethod
    def divide(num1: TnNum, num2: TnNum, digits: TnInt = None) -> TnNum:
        if not num1:
            return num2
        if not num2:
            return num1
        if isinstance(num1, (int, float)):
            if isinstance(num2, (int, float)):
                if num2 != 0:
                    if num1 != 0:
                        if not digits:
                            return num1/num2
                        return round(num1/num2, digits)
                    return num2
                return num1
            return num1
        if isinstance(num2, (int, float)):
            return num2
        return None

    @staticmethod
    def multiply(num1: TnNum, num2: TnNum, digits: TnInt = None) -> TnNum:
        if not num1 or not num2:
            return None
        if isinstance(num1, (int, float)):
            if isinstance(num2, (int, float)):
                if not digits:
                    return num1*num2
                return round(num1*num2, digits)
            return None
        return None



# coding=utf-8
from collections.abc import Callable, Iterator
from typing import Any
TyArr = list[Any]
TyCallable = Callable[..., Any]
TyTup = tuple[Any]
TyDic = dict[Any, Any]
TyAoD = list[TyDic]
TyArrTup = TyArr | TyTup
TyIterAny = Iterator[Any]

TnArr = None | TyArr
TnArrTup = None | TyArrTup
TnCallable = None | TyCallable


class Obj:
    """ Manage Object
    """
    @staticmethod
    def shrink_array(obj: Any) -> Any:
        if isinstance(obj, (tuple, list)):
            if len(obj) == 1:
                return obj[0]
        return obj

    # @classmethod
    # def verify_key_type(cls, obj: Any, sw: bool = True, pytype=type(str)) -> bool:
    #     if obj is None:
    #         return sw
    #     for item in obj:
    #         if isinstance(item, dict):
    #             for key in item.keys():
    #                 if not isinstance(key, pytype):
    #                     sw = False
    #                     print(f"key: {key} is not of type: {pytype}")
    #                     print(f"key type of key: {key} is {type(key)}")
    #             for value in item.values():
    #                 if not cls.verify_key_type(value, sw, pytype):
    #                     return False
    #     return sw

    # @classmethod
    # def verify_key_type_str(
    #         cls, obj: any, sw: bool = True):
    #     return cls.verify_key_type(obj, sw, type=(str))

    @staticmethod
    def yield_aod(obj: Any) -> TyIterAny:
        """ show objects as Array of Dictionaries
        """
        if isinstance(obj, (list, tuple)):
            # for _obj in obj:
            #     yield _obj
            yield from obj
        if isinstance(obj, (dict, str)):
            yield obj

    @staticmethod
    def sh_aod_if_arr(obj: Any, fncs=None) -> TyAoD:
        """ show objects as Array of Dictionaries
        """
        aod = []
        if fncs is None:
            for _obj in obj:
                if _obj is None:
                    continue
                if not _obj == {}:
                    continue
                aod.append(obj)
            return aod
        for _obj in obj:
            obj_new = fncs(_obj)
            if obj_new is None:
                continue
            if not obj_new == {}:
                continue
            aod.append(obj_new)
        return aod

    @staticmethod
    def sh_aod_if_dic(obj: Any, fncs: TnCallable = None) -> TyAoD:
        """ show objects as Array of Dictionaries
        """
        aod: TyAoD = []
        if fncs is None:
            obj_new = obj
        else:
            obj_new = fncs(obj)
        if obj_new is not None:
            if isinstance(obj_new, (dict)):
                if obj_new != {}:
                    aod.append(obj_new)
        return aod

    @classmethod
    def sh_aod(cls, obj: Any, fncs: TnCallable = None) -> TyAoD:
        """ show objects as Array of Dictionaries
        """
        aod = []
        if fncs is None:
            if isinstance(obj, (list, tuple)):
                for _obj in obj:
                    aod.append(_obj)
            elif isinstance(obj, (dict, str)):
                aod.append(obj)
            return aod
        if isinstance(obj, (list, tuple)):
            return cls.sh_aod_if_arr(obj, fncs)
        if isinstance(obj, dict):
            return cls.sh_aod_if_dic(obj, fncs)
        return aod

    @staticmethod
    def sh_arr(obj: Any | TyArr) -> TnArr:
        if obj is None:
            return obj
        if isinstance(obj, list):
            return obj
        return [obj]

    @staticmethod
    def to_string(obj: Any, separator: str = '.') -> str:
        if obj is None:
            return ''
        if isinstance(obj, (list, tuple)):
            return separator.join(obj)
        if isinstance(obj, str):
            return obj.strip()
        if isinstance(obj, int):
            return str(obj)
        return ''

    @staticmethod
    def sh_text(obj: Any) -> Any:
        if isinstance(obj, (list, tuple)):
            return ' '.join(obj)
        return obj

    # @classmethod
    # def flatten(cls, obj):
    #     array = []
    #     for element in obj:
    #         if isinstance(element, list):
    #             array.extend(cls.flatten(element))
    #         else:
    #             array.append(element)
    #     return array

    @classmethod
    def extract_values(cls, obj, key, **kwargs):
        arr = kwargs.get('arr', [])
        if isinstance(obj, dict):
            for k, v in obj.items():
                if isinstance(v, (dict, list)):
                    cls.extract_values(v, key, arr=arr)
                if k == key:
                    arr.append(v)
        elif isinstance(obj, list):
            for item in obj:
                cls.extract_values(item, key, arr=arr)
        return arr



# coding=utf-8
from collections.abc import Iterator
from typing import Any

TyArr = list[Any]
TnArr = None | TyArr
TyPoA = list[TnArr]
TyTup = tuple[Any, ...]

TnTup = None | TyTup


class PoA:
    """
    Manage Pair of Arrays
    """
    @staticmethod
    def yield_items(
            poa: TyPoA, obj: Any) -> Iterator[TnTup]:
        arr0 = poa[0]
        arr1 = poa[1]
        if arr0 is None:
            yield None
        elif arr1 is None:
            yield None
        else:
            for item0 in arr0:
                for item1 in arr1:
                    yield (item0, item1, obj)



# coding=utf-8
from collections.abc import Iterator
from typing import Any

import os
import datetime
import glob
import pathlib
from string import Template
import re

from ka_uts_com.log import Log

TyArr = list[Any]
TyAoS = list[str]
TyAoA = list[TyArr]
TyDic = dict[Any, Any]
TyDoA = dict[Any, TyArr]
TyDoAoA = dict[Any, TyAoA]
TyDoInt = dict[str, int]
TyDoDoInt = dict[str, TyDoInt]
TyIntStr = int | str
TyPath = str
TyIterAny = Iterator[Any]
TyStr = str
TyTup = tuple[Any]

TnArr = None | TyArr
TnAoA = None | TyAoA
TnDic = None | TyDic
TnInt = None | int
TnPath = None | TyPath
TnStr = None | str
TnTup = None | TyTup


class Path:

    @staticmethod
    def verify(path: TyPath) -> None:
        if path is None:
            raise Exception("path is None")
        elif path == '':
            raise Exception("path is empty")

    @classmethod
    def edit_path(cls, path: TyPath, kwargs: TyDic) -> str:
        _d_edit = kwargs.get('d_out_path_edit', {})
        _prefix = kwargs.get('dl_out_file_prefix', '')
        _suffix = kwargs.get('dl_out_file_suffix', '.csv')
        _edit_from = _d_edit.get('from')
        _edit_to = _d_edit.get('to')
        if _edit_from is not None and _edit_to is not None:
            _path_out = path.replace(_edit_from, _edit_to)
        else:
            _path_out = path
        _dir_out = os.path.dirname(_path_out)
        cls.mkdir_from_path(_dir_out)
        _basename_out = os.path.basename(_path_out)
        if _prefix:
            _basename_out = str(f"{_prefix}{_basename_out}")
        if _suffix:
            _basename_out = os.path.splitext(_basename_out)[0]
            _basename_out = str(f"{_basename_out}{_suffix}")
        _path_out = os.path.join(_dir_out, _basename_out)
        return _path_out

    @staticmethod
    def mkdir(path: TyPath) -> None:
        if not os.path.exists(path):
            # Create the directory
            os.makedirs(path)

    @staticmethod
    def mkdir_from_path(path: TyPath) -> None:
        _dir = os.path.dirname(path)
        if not os.path.exists(_dir):
            # Create the directory
            os.makedirs(_dir)

    @staticmethod
    def sh_basename(path: TyPath) -> str:
        """ Extracts basename of a given path.
            Should Work with any OS Path on any OS
        """
        raw_string = r'[^\\/]+(?=[\\/]?$)'
        basename = re.search(raw_string, path)
        if basename:
            return basename.group(0)
        return path

    @classmethod
    def sh_components(
        # def sh_component(
            cls, path: TyPath, d_ix: TyDic, separator: str = "-") -> TnStr:
        ix_start = d_ix.get("start")
        ix_add = d_ix.get("add", 0)
        if not ix_start:
            return None
        _a_dir: TyArr = cls.split_to_array(path)
        _ix_end = ix_start + ix_add + 1
        _component = separator.join(_a_dir[ix_start:_ix_end])
        _component = os.path.splitext(_component)[0]
        return _component

    @classmethod
    def sh_component_using_field_name(
        # def sh_component_at_start(
            cls, path: TyPath, d_path_ix: TyDoDoInt, field_name: str) -> TyStr:
        _d_ix: TyDoInt = d_path_ix.get(field_name, {})
        if not _d_ix:
            msg = f"field_name: {field_name} is not defined in dictionary: {d_path_ix}"
            raise Exception(msg)
        _start = _d_ix.get('start')
        if not _start:
            msg = f"'start' is not defined in dictionary: {_d_ix}"
            raise Exception(msg)
        _a_dir: TyAoS = cls.split_to_array(path)
        if _start < len(_a_dir):
            return _a_dir[_start]
        msg = f"index: {_start} is out of range of list: {_a_dir}"
        raise Exception(msg)

    @staticmethod
    def sh_fnc_name_using_pathlib(path: TyPath) -> str:
        # def sh_fnc_name(path: TyPath) -> str:
        _purepath = pathlib.PurePath(path)
        dir_: str = _purepath.parent.name
        stem_: str = _purepath.stem
        return f"{dir_}-{stem_}"

    @staticmethod
    def sh_fnc_name_using_os_path(path: TyPath) -> str:
        # def sh_os_fnc_name(path: TyPath) -> str:
        split_ = os.path.split(path)
        dir_ = os.path.basename(split_[0])
        stem_ = os.path.splitext(split_[1])[0]
        return f"{dir_}-{stem_}"

    @classmethod
    def sh_last_component(cls, path: TyPath) -> Any:
        a_dir: TyArr = cls.split_to_array(path)
        return a_dir[-1]

    @classmethod
    def sh_path_using_pathnm(cls, pathnm: str, **kwargs) -> str:
        # def sh_path_by_pathnm(cls, pathnm: str, **kwargs) -> str:
        _path = kwargs.get(pathnm, '')
        _dir_dat = kwargs.get('dir_dat')
        # print(f"sh_path_using_pathnm _path = {_path}")
        # print(f"sh_path_using_pathnm _dir_dat = {_dir_dat}")
        if _dir_dat is not None:
            _tenant = kwargs.get('tenant', '')
            _path = os.path.join(_dir_dat, _tenant, _path)
        _d_path = kwargs.get('d_path', {})
        _path = cls.sh_path_using_d_path(_path, _d_path)
        # print(f"sh_path_using_pathnm _path = {_path}")
        path = cls.sh_path_using_d_datetype(_path, pathnm, kwargs)
        # print(f"sh_path_using_pathnm path = {path}")
        return path

    @staticmethod
    def sh_path_using_d_path(path: TyPath, d_path: TyDic) -> str:
        # def sh_path_by_d_path(path: TyPath, d_path: TyDic) -> str:
        if not d_path:
            return path
        return Template(path).safe_substitute(d_path)

    @classmethod
    def sh_path_using_d_datetype(
        # def sh_path_by_d_pathnmh2dttype(
            cls, path: TyPath, pathnm: str, kwargs) -> str:
        _d_datetype: TyDic = kwargs.get('d_pathnm2dttype', {})
        _datetype: TyDoA = _d_datetype.get(pathnm, {})
        Log.Eq.debug("path", path)
        Log.Eq.debug("pathnm", pathnm)
        Log.Eq.debug("_d_datetype", _d_datetype)
        Log.Eq.debug("_datetype", _datetype)
        match _datetype:
            case 'last':
                path_new = cls.sh_path_last(path)
            case 'first':
                path_new = cls.sh_path_first(path)
            case 'now':
                path_new = cls.sh_path_now(path, **kwargs)
            case _:
                path_new = cls.sh_path(path)
        Log.Eq.debug("path_new", path_new)
        return path_new

    @staticmethod
    def sh_path(path: TyPath) -> str:
        Log.Eq.debug("path", path)
        if not path:
            raise Exception("Argument 'path' is empty")
        _a_path: TyArr = glob.glob(path)
        if not _a_path:
            msg = f"glob.glob find no paths for template: {path}"
            raise Exception(msg)
        path_new: str = sorted(_a_path)[0]
        return path_new

    @staticmethod
    def sh_path_first(path: TyPath) -> str:
        if not path:
            raise Exception("Argument 'path' is empty")
        _a_path: TyArr = glob.glob(path)
        if not _a_path:
            msg = f"glob.glob find no paths for template: {path}"
            raise Exception(msg)
        path_new: str = sorted(_a_path)[0]
        return path_new

    @staticmethod
    def sh_path_last(path: TyPath) -> str:
        if not path:
            raise Exception("Argument 'path' is empty")
        _a_path: TyArr = glob.glob(path)
        if not _a_path:
            msg = f"glob.glob find no paths for template: {path}"
            raise Exception(msg)
        path_new: str = sorted(_a_path)[-1]
        return path_new

    @staticmethod
    def sh_path_now(path: TyPath, **kwargs) -> str:
        now_var = kwargs.get('now_var', 'now')
        now_fmt = kwargs.get('now_fmt', '%Y%m%d')
        if not path:
            raise Exception("Argument 'path' is empty")
        _current_date: str = datetime.datetime.now().strftime(now_fmt)
        _dic = {now_var: _current_date}
        path_new: str = Template(path).safe_substitute(_dic)
        return path_new

    @staticmethod
    def split_to_array(path: TyPath) -> TyArr:
        """ Convert path to normalized pyth
            Should Work with any OS Path on any OS
        """
        normalized_path = os.path.normpath(path)
        a_path: TyArr = normalized_path.split(os.sep)
        return a_path



# coding=utf-8
from collections.abc import Callable, Iterator

import glob
import os

from ka_uts_com.log import Log
from ka_uts_com.timer import Timer
from ka_uts_com.utils.str import Str
from ka_uts_arr.aod import AoD

from typing import Any
TyAny = Any
TyArr = list[Any]
TyDic = dict[Any, Any]
TyAoD = list[TyDic]
TyDoD = dict[Any, TyDic]
TyFnc = Callable[..., Any]
TyIoS = Iterator[str]
TyObj = Any
TyPath = str
TnBool = None | bool
TnDic = None | TyDic
TnPath = None | TyPath
TnStr = None | str
TnFnc = None | TyFnc


class File:

    @staticmethod
    def count(path_pattern: TyPath) -> int:
        """
        count number of paths that match path pattern
        """
        return len(list(glob.iglob(path_pattern)))

    @staticmethod
    def ex_get_aod_using_fnc(
            path: TyPath, fnc: TyFnc, kwargs: TyDic) -> TyAoD:
        _mode = kwargs.get('mode', 'r')
        _aod: TyAoD = []
        with open(path, _mode) as _fd:
            for _line in _fd:
                _dic = Str.sh_dic(_line)
                _obj = fnc(_dic, kwargs)
                AoD.add(_aod, _obj)
        return _aod

    @staticmethod
    def ex_get_aod(path: TyPath, kwargs: TyDic) -> TyAoD:
        _mode = kwargs.get('mode', 'r')
        _aod: TyAoD = []
        with open(path, _mode) as _fd:
            for _line in _fd:
                _dic = Str.sh_dic(_line)
                AoD.add(_aod, _dic)
        return _aod

    @staticmethod
    def ex_get_dod_using_fnc(
            path: TyPath, fnc: TyFnc, key: str, kwargs: TyDic) -> TyDoD:
        """
        Create a dictionary of dictionaries by executing the following steps:
        1. read every line of the json file <path> into a dictionary
        2. If the function <obj_fnc> is not None:
             transform the dictionary by the function <obj_fnc>
             get value of the transformed dictionary for the key <key>
           else
             get value of the dictionary for the key <key>
        2. if the value is not None:
             insert the transformed dictionary into the new dictionary
             of dictionaries with the value as key.
        """
        _mode = kwargs.get('mode', 'r')
        _dod: TyDoD = {}
        with open(path, _mode) as _fd:
            for _line in _fd:
                _obj = Str.sh_dic(_line)
                _obj = fnc(_obj, kwargs)
                if _obj is not None:
                    _key = _obj.get(key)
                    if _key is not None:
                        _dod[_key] = _obj
        return _dod

    @staticmethod
    def ex_get_dod(
            path: TyPath, key: str, kwargs: TyDic) -> TyDoD:
        """
        Create a dictionary of dictionaries by executing the following steps:
        1. read every line of the json file <path> into a dictionary
        2. If the function <obj_fnc> is not None:
             transform the dictionary by the function <obj_fnc>
             get value of the transformed dictionary for the key <key>
           else
             get value of the dictionary for the key <key>
        2. if the value is not None:
             insert the transformed dictionary into the new dictionary
             of dictionaries with the value as key.
        """
        _mode = kwargs.get('mode', 'r')
        _dod: TyDoD = {}
        with open(path, _mode) as _fd:
            for _line in _fd:
                _obj = Str.sh_dic(_line)
                if _obj is not None:
                    _key = _obj.get(key)
                    _key = _obj[key]
                    if _key is not None:
                        _dod[_key] = _obj
        return _dod

    @classmethod
    def get_aod(cls, path: TyPath, fnc: TnFnc, kwargs: TyDic) -> TyAoD:
        Timer.start(cls.get_aod, f"{path}")
        if fnc is not None:
            _aod = cls.ex_get_aod_using_fnc(path, fnc, kwargs)
        else:
            _aod = cls.ex_get_aod(path, kwargs)
        Timer.end(cls.get_aod, f"{path}")
        return _aod

    @classmethod
    def get_dic(cls, path: TyPath, fnc: TnFnc, kwargs: TyDic) -> TyDic:
        Timer.start(cls.get_aod, f"{path}")
        aod = cls.get_aod(path, fnc, kwargs)
        if len(aod) > 1:
            msg = (f"File {path} contains {len(aod)} records; "
                   "it should contain only one record")
            raise Exception(msg)
        return aod[0]

    @classmethod
    def get_dod(
            cls, path: TyPath, fnc: TnFnc, key: str, kwargs: TyDic) -> TyDoD:
        """
        Create a dictionary of dictionaries by executing the following steps:
        1. read every line of the json file <path> into a dictionary
        2. If the function <obj_fnc> is not None:
             transform the dictionary by the function <obj_fnc>
             get value of the transformed dictionary for the key <key>
           else
             get value of the dictionary for the key <key>
        2. if the value is not None:
             insert the transformed dictionary into the new dictionary
             of dictionaries with the value as key.
        """
        Timer.start(cls.get_aod, f"{path}")
        if fnc is not None:
            _dod = cls.ex_get_dod_using_fnc(path, fnc, key, kwargs)
        else:
            _dod = cls.ex_get_dod(path, key, kwargs)
        Timer.end(cls.get_aod, f"{path}")
        return _dod

#   class Latest:
#
#       @staticmethod
#       def get(path_pattern: str) -> TnStr:
#           """
#           get latest path that match path pattern
#           """
#           return File.get_latest(path_pattern)

    @staticmethod
    def get_latest(path_pattern: TyPath) -> TnPath:
        """
        get latest path that match path pattern
        """
        _iter_path = glob.iglob(path_pattern)
        _a_path = list(_iter_path)
        if len(_a_path) > 0:
            return max(_a_path, key=os.path.getmtime)
        msg = f"No path exist for pattern: {path_pattern}"
        Log.error(msg)
        return None

    @staticmethod
    def get_paths(
            path_pattern: TyPath, sw_recursive: TnBool = None) -> TyIoS:
        """
        get all paths that match path_pattern
        """
        if sw_recursive is None:
            sw_recursive = False
        _paths: Iterator[str] = glob.iglob(path_pattern, recursive=sw_recursive)
        Log.Eq.debug("path_pattern", path_pattern)
        Log.Eq.debug("_paths", _paths)
        for _path in _paths:
            if os.path.isfile(_path):
                Log.Eq.debug("_path", _path)
                yield _path

    @staticmethod
    def io(obj: TyObj, path: TyPath, fnc: TyFnc) -> None:
        """
        execute io function
        """
        fnc(obj, path)



# coding=utf-8

from typing import Any
TyBytes = bytes
TyDoS = dict[Any, str]
TnDoS = None | TyDoS


class Byte:
    """ Manage Byte Class
    """
    @staticmethod
    def replace_by_dic(bytes_: TyBytes, **kwargs) -> str:
        dic: TnDoS = kwargs.get('dic_replace')
        if not dic:
            return bytes_.decode('utf-8')
        string = bytes_.decode('utf-8')
        for k, v in dic.items():
            string = string.replace(k, v)
        return string
